from collections import deque
from pathlib import Path
from typing import IO

import numpy


MAZE_SIZE = 16

VISITED_BIT = 1
EAST_BIT = 2
SOUTH_BIT = 4
WEST_BIT = 8
NORTH_BIT = 16


def read_walls(walls, x, y, direction):
    if walls is None:
        return (0, 0, 0)
    wall = walls[x][y]
    detections = deque([
        wall & EAST_BIT,
        wall & SOUTH_BIT,
        wall & WEST_BIT,
        wall & NORTH_BIT,
    ])
    if direction == 'E':
        detections.rotate(1)
    elif direction == 'N':
        detections.rotate(2)
    elif direction == 'W':
        detections.rotate(-1)
    elif direction == 'S':
        pass
    return tuple([bool(x) for x in detections][:3])


def _read_maze_oshwdem(txt: str) -> numpy.ndarray:
    txt = '\n'.join(txt.splitlines()[1:])
    txt = [''.join(i) for i in zip(*txt.splitlines())]

    south = txt[::4][:-1]
    south = [[1 if wall == '|' else 0
              for wall in row.replace('B', '+').strip('+').split('+')]
             for row in south]
    south = numpy.array(south).astype('uint8').T
    north = numpy.roll(south, -1, axis=1)

    west = txt[1::4]
    west = [[1 if wall == '-' else 0
             for wall in column[::2][:-1]]
            for column in west]
    west = numpy.array(west).astype('uint8').T
    east = numpy.roll(west, -1, axis=0)

    east *= EAST_BIT
    south *= SOUTH_BIT
    west *= WEST_BIT
    north *= NORTH_BIT
    return east + south + west + north


def _read_maze_default(txt: str) -> numpy.ndarray:
    txt = txt.splitlines()
    post_char = txt[0][0]

    south = txt[::2][1:][::-1]
    south = [[1 if wall == '---' else 0
              for wall in row.strip(post_char).split(post_char)]
             for row in south]
    south = numpy.array(south).astype('uint8').T
    north = numpy.roll(south, -1, axis=1)

    west = txt[1::2][::-1]
    west = [[1 if wall == '|' else 0
             for wall in column[::4][:-1]]
            for column in west]
    west = numpy.array(west).astype('uint8').T
    east = numpy.roll(west, -1, axis=0)

    east *= EAST_BIT
    south *= SOUTH_BIT
    west *= WEST_BIT
    north *= NORTH_BIT
    return east + south + west + north


def load_maze(data: IO) -> numpy.ndarray:
    if isinstance(data, Path):
        maze = data.read_text()
    else:
        maze = data.read()
    if maze.startswith('OSHWDEM'):
        return _read_maze_oshwdem(maze)
    return _read_maze_default(maze)



from pathlib import Path
import struct
import sys
import zmq

from PyQt5 import QtCore, QtWidgets
from PyQt5.QtWidgets import QLineEdit
from PyQt5.QtWidgets import QListWidget
from PyQt5.QtWidgets import QSlider
from PyQt5.QtWidgets import QSplitter
from PyQt5.QtWidgets import QStatusBar
from PyQt5.QtWidgets import QVBoxLayout
from PyQt5.QtWidgets import QWidget
from pyqtgraph import GraphicsLayoutWidget

from .mazes import load_maze
from .graphics import MazeItem


class ZMQListener(QtCore.QObject):

    message = QtCore.pyqtSignal(bytes)

    def __init__(self, context, host, port):
        super().__init__()

        self.rep = context.socket(zmq.REP)
        self.rep.bind('tcp://{host}:{port}'.format(host=host, port=port))
        self.pull = context.socket(zmq.PULL)
        self.pull.bind('inproc://reply')

        self.poller = zmq.Poller()
        self.poller.register(self.rep, zmq.POLLIN)
        self.poller.register(self.pull, zmq.POLLIN)

        self.running = True

    def loop(self):
        while self.running:
            events = dict(self.poller.poll(10))
            if not events:
                continue
            self.process_events(events)

    def process_events(self, events):
        for socket in events:
            if events[socket] != zmq.POLLIN:
                continue
            if socket == self.rep:
                self.message.emit(socket.recv())
            elif socket == self.pull:
                self.rep.send(socket.recv())


class MainWindow(QtWidgets.QMainWindow):
    def __init__(self, host, port, path, parent=None):
        super().__init__(parent)

        self.path = path
        self.maze_files = sorted(
            fname.relative_to(self.path)
            for fname in self.path.glob('**/*.txt') if fname.is_file()
        )

        self.setWindowTitle('Micromouse maze simulator')
        self.resize(800, 600)

        self.history = []

        self.status = QStatusBar()
        self.setStatusBar(self.status)

        self.search = QLineEdit()
        self.search.textChanged.connect(self.filter_mazes)

        self.files = QListWidget()
        self.files.currentItemChanged.connect(self.list_value_changed)

        self.graphics = GraphicsLayoutWidget()
        viewbox = self.graphics.addViewBox()
        viewbox.setAspectLocked()
        self.maze = MazeItem()
        viewbox.addItem(self.maze)

        self.slider = QSlider(QtCore.Qt.Horizontal)
        self.slider.setSingleStep(1)
        self.slider.setPageStep(10)
        self.slider.setTickPosition(QSlider.TicksAbove)
        self.slider.valueChanged.connect(self.slider_value_changed)
        self.reset()

        files_layout = QVBoxLayout()
        files_layout.setContentsMargins(0, 0, 0, 0)
        files_layout.addWidget(self.search)
        files_layout.addWidget(self.files)
        files_widget = QWidget()
        files_widget.setLayout(files_layout)
        graphics_layout = QVBoxLayout()
        graphics_layout.setContentsMargins(0, 0, 0, 0)
        graphics_layout.addWidget(self.graphics)
        graphics_layout.addWidget(self.slider)
        graphics_widget = QWidget()
        graphics_widget.setLayout(graphics_layout)
        central_splitter = QSplitter()
        central_splitter.addWidget(files_widget)
        central_splitter.addWidget(graphics_widget)

        main_layout = QVBoxLayout()
        main_layout.addWidget(central_splitter)
        main_layout.setContentsMargins(4, 4, 4, 4)
        main_widget = QWidget()
        main_widget.setLayout(main_layout)

        self.setCentralWidget(main_widget)
        self.filter_mazes('')

        self.context = zmq.Context()
        self.reply = self.context.socket(zmq.PUSH)
        self.reply.connect('inproc://reply')

        self.thread = QtCore.QThread()
        self.zeromq_listener = ZMQListener(self.context, host=host, port=port)
        self.zeromq_listener.moveToThread(self.thread)

        self.thread.started.connect(self.zeromq_listener.loop)
        self.zeromq_listener.message.connect(self.signal_received)

        QtCore.QTimer.singleShot(0, self.thread.start)

    def filter_mazes(self, text):
        keywords = text.lower().split(' ')
        self.files.clear()
        for fname in self.maze_files:
            for key in keywords:
                if key not in str(fname).lower():
                    break
            else:
                self.files.addItem(str(fname))
        self.files.setCurrentRow(0)

    def list_value_changed(self, after, before):
        if not after:
            return
        self.set_maze(after.text())

    def set_maze(self, fname):
        template_file = Path(fname)
        template = load_maze(self.path / template_file)
        self.maze.reset(template)
        self.reset()

    def reset(self):
        self.history = []
        self.slider.setValue(-1)
        self.slider.setRange(-1, -1)
        self.status.showMessage('Ready')

    def slider_update(self):
        self.slider.setTickInterval(len(self.history) / 10)
        self.slider.setRange(0, len(self.history) - 1)
        self.status_set_slider(self.slider.value())

    def slider_value_changed(self, value):
        self.status_set_slider(value)
        if not len(self.history):
            return
        state = self.history[value]
        position = state[:3]
        discovery = state[3:]
        self.maze.update_position(position)
        self.maze.update_discovery(discovery)

    def status_set_slider(self, value):
        self.status.showMessage('{}/{}'.format(value, len(self.history) - 1))

    def signal_received(self, message):
        if message.startswith(b'W'):
            walls = self.maze.read_position_walls(message.lstrip(b'W'))
            self.reply.send(struct.pack('3B', *walls))
            return
        if message.startswith(b'S'):
            message = message.lstrip(b'S')
            self.history.append(message)
            self.reply.send(b'ok')
            self.slider_update()
            return
        if message == b'reset':
            self.reset()
            self.reply.send(b'ok')
            return
        if message == b'ping':
            self.reply.send(b'pong')
            return
        raise ValueError('Unknown message received! "{}"'.format(message))

    def closeEvent(self, event):
        self.zeromq_listener.running = False
        self.thread.quit()
        self.thread.wait()


def run(host, port, path):
    app = QtWidgets.QApplication(sys.argv)
    main = MainWindow(host=host, port=port, path=path)
    main.show()
    sys.exit(app.exec_())



import os
import tarfile
from pathlib import Path
from urllib.request import urlretrieve

from .typing import TarMembers


def select_tar_members(
        members: TarMembers, parent: Path, new_path: Path) -> TarMembers:
    """
    Select .tar members given a parent path and rename them to have a new
    path instead.

    Parameters
    ----------
    members
        A list of .tar members.
    parent
        Parent path to filter members.
    new_path
        New member path.

    Returns
    -------
    A clean .tar member list with the selected, renamed components.
    """
    clean = []
    for member in members:
        path = Path(member.name)
        if path.is_absolute():
            raise NotImplementedError('Please, report this unexpected error!')
        path = path.relative_to(path.parts[0])
        if path.parent != parent:
            continue
        member.name = str(new_path / path.name)
        clean.append(member)
    return clean


def download_micromouseonline_mazes(download_path: Path):
    """
    Download Micromouseonline mazes.

    Parameters
    ----------
    download_path
        Where to download the maze files to.
    """
    url = 'https://github.com/micromouseonline/micromouse_maze_tool/'\
        'archive/master.tar.gz'
    reveal_tar, headers = urlretrieve(url)
    with tarfile.open(reveal_tar) as tar:
        members = select_tar_members(
            tar.getmembers(),
            parent=Path('mazefiles/text'),
            new_path=Path('16x16')
        )
        tar.extractall(str(download_path), members)
    os.remove(reveal_tar)



from pathlib import Path

import click

from .ui import run
from .download import download_micromouseonline_mazes


@click.command()
@click.argument('mazes_path', type=click.Path(), default=Path.home()/'.mmsim')
@click.option('-h', '--host', type=str, default='127.0.0.1',
              help='Listen on IP (default: 127.0.0.1).')
@click.option('-p', '--port', type=int, default=6574,
              help='Listen on port (default: 6574).')
def launch(mazes_path: Path, host: str='127.0.0.1', port: int=6574):
    """
    Launch the Micromouse Maze Simulator interface.
    """
    mazes_path = Path(mazes_path)
    if not mazes_path.exists():
        download_micromouseonline_mazes(mazes_path)
    run(host, port, Path(mazes_path))



from itertools import product
import struct

import numpy
from pyqtgraph import (
    mkBrush,
    mkPen,
    GraphicsObject,
    QtCore,
    QtGui,
)

from .mazes import read_walls

from .mazes import EAST_BIT
from .mazes import SOUTH_BIT
from .mazes import WEST_BIT
from .mazes import NORTH_BIT
from .mazes import VISITED_BIT

from .mazes import MAZE_SIZE


CELL_WIDTH = 180
WALL_WIDTH = 12

GRAY = (100, 100, 100)
GREEN = (0, 255, 0)
RED = (255, 0, 0)
WHITE = (255, 255, 255)


def paint_walls(painter, walls, color):
    painter.setBrush(mkBrush(color))
    painter.setPen(mkPen(None))
    for (x, y) in product(range(MAZE_SIZE), repeat=2):
        wall = walls[y][x]
        if wall & EAST_BIT:
            painter.drawRect(QtCore.QRectF(
                (y + 1) * CELL_WIDTH - WALL_WIDTH / 2,
                -(x + 1) * CELL_WIDTH + WALL_WIDTH / 2,
                WALL_WIDTH,
                CELL_WIDTH,
            ))
        if wall & SOUTH_BIT:
            painter.drawRect(QtCore.QRectF(
                y * CELL_WIDTH + WALL_WIDTH / 2,
                -x * CELL_WIDTH + WALL_WIDTH / 2,
                CELL_WIDTH,
                WALL_WIDTH,
            ))
        if wall & WEST_BIT:
            painter.drawRect(QtCore.QRectF(
                y * CELL_WIDTH - WALL_WIDTH / 2,
                -(x + 1) * CELL_WIDTH + WALL_WIDTH / 2,
                WALL_WIDTH,
                CELL_WIDTH,
            ))
        if wall & NORTH_BIT:
            painter.drawRect(QtCore.QRectF(
                y * CELL_WIDTH + WALL_WIDTH / 2,
                -(x + 1) * CELL_WIDTH + WALL_WIDTH / 2,
                CELL_WIDTH,
                WALL_WIDTH,
            ))


def paint_discovered(painter, distances, walls):
    if walls is not None:
        paint_walls(painter, walls, color=WHITE)
    for (x, y) in product(range(MAZE_SIZE), repeat=2):
        painter.setPen(mkPen(color=GRAY))
        if walls is not None:
            wall = walls[x][y]
            if wall & VISITED_BIT:
                painter.setPen(mkPen(color=GREEN))
        painter.drawText(QtCore.QRectF(
            (x + .5) * CELL_WIDTH + WALL_WIDTH / 2 - 50,
            -(y + .5) * CELL_WIDTH + WALL_WIDTH / 2 - 50,
            100, 100),
            QtCore.Qt.AlignCenter,
            "%s" % distances[x][y])


def paint_template(painter, walls):
    if walls is not None:
        paint_walls(painter=painter, walls=walls, color=GRAY)
    for (x, y) in product(range(MAZE_SIZE + 1), repeat=2):
        painter.setBrush(mkBrush(WHITE))
        painter.setPen(mkPen(None))
        painter.drawRect(QtCore.QRectF(
            x * CELL_WIDTH - WALL_WIDTH / 2,
            -y * CELL_WIDTH + WALL_WIDTH / 2,
            WALL_WIDTH,
            WALL_WIDTH,
        ))


def paint_position(painter, x, y, direction):
    painter.setBrush(mkBrush(RED))
    painter.setPen(mkPen(None))
    if direction in ['E', 'W']:
        robot_width = 50
        robot_height = 100
    else:
        robot_width = 100
        robot_height = 50
    x_compensation = 0
    y_compensation = 0
    if direction == 'E':
        y_compensation = robot_height / 2
    elif direction == 'S':
        x_compensation = (CELL_WIDTH - robot_width) / 2
    elif direction == 'W':
        x_compensation = CELL_WIDTH - robot_width
        y_compensation = robot_height / 2
    elif direction == 'N':
        x_compensation = (CELL_WIDTH - robot_width) / 2
        y_compensation = CELL_WIDTH - robot_height
    painter.drawRect(QtCore.QRectF(
        x * CELL_WIDTH + x_compensation,
        -(y + 1) * CELL_WIDTH + y_compensation,
        robot_width,
        robot_height,
    ))


class MazeItem(GraphicsObject):
    def __init__(self):
        super().__init__()
        self.reset(None)

    def reset(self, template):
        self.distances = None
        self.walls = None
        self.template = template
        self.x = 0
        self.y = 0
        self.direction = 0

        self.picture = QtGui.QPicture()
        self.position_picture = QtGui.QPicture()

        self.generateTemplate()
        self.update()

    def generateTemplate(self):
        self.template_picture = QtGui.QPicture()
        painter = QtGui.QPainter(self.template_picture)
        painter.scale(1, -1)
        paint_template(painter=painter, walls=self.template)
        painter.end()

    def generatePicture(self):
        self.picture = QtGui.QPicture()
        painter = QtGui.QPainter(self.picture)
        painter.setFont(QtGui.QFont('times', 50))
        painter.scale(1, -1)
        paint_discovered(painter, distances=self.distances, walls=self.walls)
        painter.end()

    def generatePosition(self):
        self.position_picture = QtGui.QPicture()
        painter = QtGui.QPainter(self.position_picture)
        painter.scale(1, -1)
        paint_position(painter, x=self.x, y=self.y, direction=self.direction)
        painter.end()

    def paint(self, p, *args):
        p.drawPicture(0, 0, self.template_picture)
        p.drawPicture(0, 0, self.position_picture)
        p.drawPicture(0, 0, self.picture)

    def boundingRect(self):
        return QtCore.QRectF(self.template_picture.boundingRect())

    def update_position(self, position):
        self.x, self.y, self.direction = struct.unpack('3B', position)
        self.direction = chr(self.direction)
        self.generatePosition()
        self.update()
        return read_walls(self.template, self.x, self.y, self.direction)

    def read_position_walls(self, position):
        x, y, direction = struct.unpack('3B', position)
        return read_walls(self.template, x, y, chr(direction))

    def update_discovery(self, discovery):
        order = chr(discovery[0])
        distances = discovery[1:257]
        distances = numpy.frombuffer(distances, dtype='uint8')
        distances = distances.reshape(MAZE_SIZE, MAZE_SIZE).T
        if order == 'F':
            distances = distances.T

        order = chr(discovery[257])
        walls = discovery[258:]
        walls = numpy.frombuffer(walls, dtype='uint8')
        walls = walls.reshape(MAZE_SIZE, MAZE_SIZE).T
        if order == 'F':
            walls = walls.T

        self.walls = walls
        self.distances = distances
        self.generatePicture()
        self.update()



"""
Setup module.
"""
from setuptools import setup
from mmsim import __version__


setup(
    name='mmsim',
    version=__version__,
    description='A simple Micromouse Maze Simulator server',
    long_description='''The server can load different mazes and any client
        can connect to it to ask for the current position walls, move from
        one cell to another and visualize the simulated micromouse state.''',
    url='https://github.com/Theseus/mmsim',
    author='Miguel Sánchez de León Peque',
    author_email='peque@neosit.es',
    license='BSD License',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Topic :: Utilities',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    keywords='micromouse maze server simulator',
    entry_points={
        'console_scripts': [
            'mmsim = mmsim.commands:launch',
        ],
    },
    packages=['mmsim'],
    install_requires=[
        'click',
        'numpy',
        'pyqtgraph',
        'pyqt5',
        'pyzmq'],
    extras_require={
        'dev': [],
        'test': ['tox'],
        'docs': ['sphinx', 'sphinx_rtd_theme', 'doc8'],
    },
)
